@@ -1,6 +1,7 @@
 from .flow import Flow, FlowBuilder  # noqa: F401
 from .decorators import (  # noqa: F401
     version,
+    version_no_warnings,
     output,
     outputs,
     docs,
@@ -12,10 +13,10 @@ from .decorators import (  # noqa: F401
     changes_per_run,
     accepts,
     returns,
-    aip_task_config,
+    run_in_aip,
 )
 
 from . import protocol  # noqa: F401
 from . import util  # noqa: F401
 
-__version__ = u"0.9.2"
+__version__ = u"0.10.0"
