@@ -91,14 +91,43 @@ def dot_from_graph(graph, vertical=False, curvy_lines=False, name=None):
         rankdir="TB" if vertical else "LR",
     )
 
-    node_lists_by_cluster = defaultdict(list)
+    # First, we cluster together any nodes that share an entity name. This includes
+    # nodes generated by a common tuple function, and nodes that differ only by case
+    # key.
+    node_clusters_by_entity_name = defaultdict(list)
     for node in graph.nodes():
-        descriptor = graph.nodes[node]["descriptor"]
-        node_lists_by_cluster[descriptor].append(node)
+        entity_names = list(node.dnode.all_entity_names())
+        # If this node has no entity names at all, we'll pretend it has a single entity
+        # name. (I don't think this can currently happen, because the only descriptor
+        # with no entity names is the empty tuple `()`, and specifying an empty tuple
+        # as an output descriptor ends up doing nothing. Still, we'll try to handle this
+        # gracefully in case things change.)
+        if len(entity_names) == 0:
+            entity_names = [""]
 
-    descriptors = list(set(graph.nodes[node]["descriptor"] for node in graph.nodes()))
-    color_strs_by_descriptor = hpluv_color_dict(
-        descriptors, saturation=99, lightness=90
+        # If this node has multiple entity names, we want to make sure all those names
+        # are associated with the same cluster, so we'll merge all the clusters we find.
+        first_cluster = node_clusters_by_entity_name[entity_names[0]]
+        for other_entity_name in entity_names[1:]:
+            other_cluster = node_clusters_by_entity_name[other_entity_name]
+            if other_cluster is not first_cluster:
+                first_cluster.extend(other_cluster)
+                node_clusters_by_entity_name[other_entity_name] = first_cluster
+
+        first_cluster.append(node)
+
+    # Now we deduplicate the clusters, arrange them in a deterministic order, and assign
+    # a color to each one. (The determinism is important so that the colored graph looks
+    # the same each time.)
+    unique_node_clusters_by_obj_id = {
+        id(cluster): cluster for cluster in node_clusters_by_entity_name.values()
+    }
+    sorted_node_clusters = sorted(unique_node_clusters_by_obj_id.values(), key=min)
+    cluster_ixs = list(range(len(sorted_node_clusters)))
+    color_strs_by_cluster_ix = hpluv_color_dict(
+        cluster_ixs,
+        saturation=99,
+        lightness=90,
     )
 
     def name_from_node(node):
@@ -110,22 +139,18 @@ def dot_from_graph(graph, vertical=False, curvy_lines=False, name=None):
     def doc_from_node(node):
         return graph.nodes[node].get("doc")
 
-    for cluster_ix, node_list in enumerate(node_lists_by_cluster.values()):
-        sorted_nodes = list(
-            sorted(node_list, key=lambda node: graph.nodes[node]["task_ix"])
-        )
-
-        # We use a numerical cluster index instead of the actual cluster descriptor,
+    for cluster_ix, node_cluster in zip(cluster_ixs, sorted_node_clusters):
+        # We use a numerical cluster index instead of something like an entity name
         # because pydot can break if the cluster name constains special characters.
         subdot = pydot.Cluster(str(cluster_ix), style="invis")
+        color = color_strs_by_cluster_ix[cluster_ix]
 
-        for node in sorted_nodes:
-            descriptor = graph.nodes[node]["descriptor"]
+        for node in sorted(node_cluster):
             doc = doc_from_node(node)
             dot_node = pydot.Node(
                 name_from_node(node),
                 style="filled",
-                fillcolor=color_strs_by_descriptor[descriptor],
+                fillcolor=color,
                 shape="box",
             )
             if doc:
