@@ -8,12 +8,17 @@ import warnings
 
 import attr
 
-from .datatypes import ResultGroup, DescriptorMetadata, TaskKey
+from .datatypes import ResultGroup, DescriptorMetadata, TaskKey, VersioningPolicy
 from .descriptors.parsing import entity_dnode_from_descriptor
 from .descriptors import ast
 from .deps.optdep import import_optional_dependency
 from .exception import UndefinedEntityError
-from .core.flow_execution import TaskCompletionRunner, TaskKeyLogger
+from .core.flow_execution import (
+    ExecutionContext,
+    MemoryResultCache,
+    TaskCompletionRunner,
+    TaskKeyLogger,
+)
 from .core.task_execution import TaskState
 from .protocols import TupleProtocol, NonSerializableObjectProtocol
 from .provider import (
@@ -45,8 +50,8 @@ class EntityDeriver:
 
     # --- Public API.
 
-    def __init__(self, flow_state, flow_instance_uuid):
-        self._flow_state = flow_state
+    def __init__(self, flow_config, flow_instance_uuid):
+        self._flow_config = flow_config
         self._flow_instance_uuid = flow_instance_uuid
 
         # For many dnodes, we can precompute the appropriate provider ahead of time.
@@ -63,9 +68,12 @@ class EntityDeriver:
         # Tracks whether we've pre-validated the base descriptors in this flow.
         self._base_prevalidation_is_complete = False
 
-        # The "bootstrap state" needs to be complete before we can compute user-defined
-        # entities.
-        self._bootstrap = None
+        # We need an ExecutionCore object to do any derivation, but creating the core
+        # itself requires derivation. We'll use this default core to get us through the
+        # initial bootstrapping phase, and then replace it with a properly configured
+        # core.
+        self._core_is_final = False
+        self._core = BOOTSTRAP_CORE
 
     # TODO We should adjust the wording of the docstring below or refactor this a
     # little. It's not necessary for *the user* to call this method, but it is necessary
@@ -77,7 +85,7 @@ class EntityDeriver:
         necessary but allows errors to surface earlier.
         """
         self._register_static_providers()
-        self._set_up_bootstrap()
+        self._set_up_final_core()
         self._prevalidate_base_dnodes()
 
     def derive(self, dnode):
@@ -293,7 +301,7 @@ class EntityDeriver:
             else:
                 dnode.fail_match()
 
-        for provider in set(self._flow_state.providers_by_name.values()):
+        for provider in set(self._flow_config.providers_by_name.values()):
             register_provider(provider)
 
         self._static_providers_by_dnode = static_providers_by_dnode
@@ -301,28 +309,32 @@ class EntityDeriver:
         # behavior downstream.
         self._followup_dnode_lists_by_dnode = dict(followup_dnode_lists_by_dnode)
 
-    def _set_up_bootstrap(self):
+    def _set_up_final_core(self):
         """
         Initializes some key objects needed to compute user-defined entities.
         """
 
-        if self._bootstrap is not None:
+        if self._core_is_final:
             return
 
-        self._bootstrap = Bootstrap(
-            persistent_cache=self._bootstrap_singleton_entity("core__persistent_cache"),
-            versioning_policy=self._bootstrap_singleton_entity(
-                "core__versioning_policy"
-            ),
-            aip_executor=self._bootstrap_singleton_entity("core__aip_executor"),
-            process_executor=self._bootstrap_singleton_entity("core__process_executor"),
-            should_memoize_default=self._bootstrap_singleton_entity(
+        self._core = ExecutionCore(
+            persistent_cache=self._compute_core_entity("core__persistent_cache"),
+            versioning_policy=self._compute_core_entity("core__versioning_policy"),
+            aip_executor=self._compute_core_entity("core__aip_executor"),
+            process_executor=self._compute_core_entity("core__process_executor"),
+            gcs_fs=self._compute_core_entity("core__persistent_cache__gcs__fs"),
+            should_memoize_default=self._compute_core_entity(
                 "core__memoize_by_default"
             ),
-            should_persist_default=self._bootstrap_singleton_entity(
+            should_persist_default=self._compute_core_entity(
                 "core__persist_by_default"
             ),
+            should_memoize_for_query_if_uncached=self._compute_core_entity(
+                "core__temp_memoize_if_uncached"
+            ),
+            task_key_logging_level=logging.INFO,
         )
+        self._core_is_final = True
 
     def _prevalidate_base_dnodes(self):
         """
@@ -362,7 +374,7 @@ class EntityDeriver:
 
         return [
             entity_dnode_from_descriptor(entity_name)
-            for entity_name in self._flow_state.entity_defs_by_name.keys()
+            for entity_name in self._flow_config.entity_defs_by_name.keys()
         ]
 
     def _obtain_provider_for_dnode(self, dnode):
@@ -377,7 +389,7 @@ class EntityDeriver:
 
         if dnode.is_entity():
             entity_name = dnode.assume_entity().name
-            if entity_name in self._flow_state.entity_defs_by_name:
+            if entity_name in self._flow_config.entity_defs_by_name:
                 message = f"""
                 Unexpected failed to find a static provider for defined entity
                 {entity_name!r};
@@ -406,33 +418,27 @@ class EntityDeriver:
         """
 
         if dnode.is_entity():
-            entity_def = self._flow_state.get_entity_def(dnode.to_descriptor())
+            entity_def = self._flow_config.get_entity_def(dnode.to_descriptor())
 
-            # TODO It's a little gross that our metadata object depends on whether
-            # it's requested before or after we construct the bootstrap. It might be
-            # nice if we could explicitly mark each entity with whether it's supposed
-            # to be configured before or after the bootstrap. However, that's
+            # TODO It's a little gross that our metadata object depends on self._core,
+            # which is different depending on whether we're in the bootstrapping phase.
+            # It might be nice if we could explicitly mark each entity with whether
+            # it's supposed to be configured before or after the core. However, that's
             # somewhat complicated by the fact that defining an entity with @builder
             # resets its entity configuration (which was probably a bad design choice
             # that we should change at some point).
-            if self._bootstrap is not None:
-                should_memoize_default = self._bootstrap.should_memoize_default
-                should_persist_default = self._bootstrap.should_persist_default
-            else:
-                should_memoize_default = True
-                should_persist_default = False
 
             if entity_def.optional_should_memoize is not None:
                 should_memoize = entity_def.optional_should_memoize
             else:
-                should_memoize = should_memoize_default
+                should_memoize = self._core.should_memoize_default
             if entity_def.needs_caching and not should_memoize:
                 # TODO Here we require that all non-deterministic values be memoized,
                 # but it would probably also be okay if they were persisted instead; we
                 # could change this check to only trigger if persistence is not
                 # enabled.
                 descriptor = dnode.to_descriptor()
-                if should_memoize_default:
+                if self._core.should_memoize_default:
                     fix_message = (
                         "removing `memoize(False)` from the corresponding function"
                     )
@@ -453,12 +459,12 @@ class EntityDeriver:
             if entity_def.optional_should_persist is not None:
                 should_persist = entity_def.optional_should_persist
             else:
-                should_persist = should_persist_default
-            if should_persist and self._bootstrap is None:
+                should_persist = self._core.should_persist_default
+            if should_persist and self._core.persistent_cache is None:
                 descriptor = dnode.to_descriptor()
                 message = f"""
                 Descriptor {descriptor!r} is set to be persisted, but it can't be
-                because core bootstrap entities depend on it.
+                because core entities depend on it.
                 The corresponding value will not be serialized and deserialized,
                 which may cause that value to be subtly different. To avoid this
                 warning, disable persistence by applying `@persist(False)` or
@@ -470,10 +476,16 @@ class EntityDeriver:
                 logger.warn(oneline(message))
                 should_persist = False
 
+            should_memoize_for_query = (
+                self._core.should_memoize_for_query_if_uncached
+                and not (should_memoize or should_persist)
+            )
+
             return DescriptorMetadata(
                 protocol=entity_def.protocol,
                 doc=entity_def.doc,
                 should_memoize=should_memoize,
+                should_memoize_for_query=should_memoize_for_query,
                 should_persist=should_persist,
             )
 
@@ -493,6 +505,7 @@ class EntityDeriver:
             return DescriptorMetadata(
                 protocol=TupleProtocol(len(dnode.children)),
                 doc=f"A Python tuple with {len(dnode.children)} values.",
+                is_composite=True,
             )
 
         else:
@@ -572,7 +585,7 @@ class EntityDeriver:
         # (code_fingerprint and changes_per_run), so we could extract that into a
         # separate class and attach it either to the TaskState or to the Task itself.
         provider = self._obtain_provider_for_dnode(dnode)
-        func_attrs = provider.get_func_attrs(case_key)
+        func_attrs = provider.get_func_attrs(case_key, self._core.versioning_policy)
 
         metadata = self._obtain_metadata_for_dnode(task.key.dnode)
 
@@ -621,11 +634,11 @@ class EntityDeriver:
         self._saved_task_states_by_key[task.key] = task_state
         return task_state
 
-    def _bootstrap_singleton_entity(self, entity_name):
+    def _compute_core_entity(self, entity_name):
         """
-        Computes the value of a 'bootstrap' entity -- i.e., a fundamental
-        internal entity needed to compute user-defined entities. Assumes the entity
-        has a single value.
+        Computes the value of a 'core' entity -- i.e., a fundamental internal entity
+        needed to compute user-defined entities. Assumes the entity has a single
+        value.
         """
 
         dnode = entity_dnode_from_descriptor(entity_name)
@@ -634,8 +647,7 @@ class EntityDeriver:
             raise ValueError(
                 oneline(
                     f"""
-                No values were defined for internal bootstrap entity
-                {entity_name!r}"""
+                No values were defined for core entity {entity_name!r}"""
                 )
             )
         if len(result_group) > 1:
@@ -643,7 +655,7 @@ class EntityDeriver:
             raise ValueError(
                 oneline(
                     f"""
-                Bootstrap entity {entity_name!r} must have exactly one
+                Core entity {entity_name!r} must have exactly one
                 value; got {len(values)} ({values!r})"""
                 )
             )
@@ -652,7 +664,7 @@ class EntityDeriver:
             raise ValueError(
                 oneline(
                     f"""
-                Bootstrap entity {entity_name!r} could not be computed because
+                Core entity {entity_name!r} could not be computed because
                 the following entities are declared but not set:
                 {", ".join(result.task_key.case_key.missing_names)}
                 """
@@ -671,12 +683,13 @@ class EntityDeriver:
             self._get_or_create_task_state_for_key(task.key) for task in dinfo.tasks
         ]
 
-        task_key_logger = TaskKeyLogger(self._bootstrap)
-        task_runner = TaskCompletionRunner(
-            bootstrap=self._bootstrap,
+        exec_context = ExecutionContext(
             flow_instance_uuid=self._flow_instance_uuid,
-            task_key_logger=task_key_logger,
+            core=self._core,
+            task_key_logger=TaskKeyLogger(self._core),
+            temp_result_cache=MemoryResultCache(),
         )
+        task_runner = TaskCompletionRunner(exec_context)
         results = task_runner.run(requested_task_states)
         assert len(results) == len(requested_task_states)
 
@@ -684,22 +697,46 @@ class EntityDeriver:
 
 
 @attr.s(frozen=True)
-class Bootstrap:
+class ExecutionCore:
     """
-    Set of entity values needed to compute / load other entities.
+    A collection of parameters and services used to compute entity values.
+
+    Most of components are themselves defined as Bionic entities; those entities are
+    computed during a special "bootstrapping" phase, in which these components are each
+    replaced with a default value.
     """
 
     persistent_cache = attr.ib()
     versioning_policy = attr.ib()
     aip_executor = attr.ib()
     process_executor = attr.ib()
-    should_memoize_default = attr.ib()
-    should_persist_default = attr.ib()
+    gcs_fs = attr.ib()
+    should_memoize_default = attr.ib(type=bool)
+    should_persist_default = attr.ib(type=bool)
+    should_memoize_for_query_if_uncached = attr.ib(type=bool)
+    task_key_logging_level = attr.ib()
 
     def evolve(self, **kwargs):
         return attr.evolve(self, **kwargs)
 
 
+BOOTSTRAP_CORE = ExecutionCore(
+    persistent_cache=None,
+    versioning_policy=VersioningPolicy(
+        check_for_bytecode_errors=False,
+        treat_bytecode_as_functional=False,
+        ignore_bytecode_exceptions=True,
+    ),
+    aip_executor=None,
+    process_executor=None,
+    gcs_fs=None,
+    should_memoize_default=True,
+    should_persist_default=False,
+    should_memoize_for_query_if_uncached=False,
+    task_key_logging_level=logging.DEBUG,
+)
+
+
 @attr.s(frozen=True)
 class DescriptorInfo:
     """
