@@ -6,16 +6,19 @@ work in a parallel setting.
 """
 
 import copy
-from functools import partial
 import logging
-from multiprocessing.managers import SyncManager
 import queue
+import re
 import sys
 import threading
 import traceback
+from concurrent.futures.thread import ThreadPoolExecutor
+from functools import partial
+from multiprocessing.managers import SyncManager
+from typing import Callable
 from uuid import uuid4
 
-from .aip.task import Task
+from .aip.task import Task, Config
 from .deps.optdep import import_optional_dependency
 from .utils.misc import oneline, SynchronizedSet
 
@@ -26,22 +29,39 @@ class AipExecutor:
     in one place.
     """
 
-    def __init__(self, aip_config):
-        self._aip_config = aip_config
+    NON_ALPHANUMERIC_PATTERN = re.compile(r"\W")
 
-    def submit(self, task_config, fn, *args, **kwargs):
-        return Task(
-            # TODO: Use a better identifiable name, maybe a combination
-            # of entity name and case key.
-            # This is a temporary workaround. We are changing the random
-            # UUID in such a manner because AIP names have to start with
-            # a letter and only accepts alphanumeric and underscore
-            # characters.
-            name="a" + str(uuid4()).replace("-", ""),
+    def __init__(
+        self,
+        gcs_fs,
+        aip_client,
+        aip_config: Config,
+        docker_image_uri_func: Callable[[], str],
+    ):
+        self._gcs_fs = gcs_fs
+        self._aip_client = aip_client
+        self._aip_config = aip_config
+        self._docker_image_uri_func = docker_image_uri_func
+
+    def submit(self, task_key, task_config, fn, *args, **kwargs):
+        task = Task(
+            name=self._create_job_name(task_key),
+            docker_image_uri=self._docker_image_uri_func(),
             config=self._aip_config,
             task_config=task_config,
             function=partial(fn, *args, **kwargs),
-        ).submit()
+        )
+
+        task.submit(gcs_fs=self._gcs_fs, aip_client=self._aip_client)
+        return ThreadPoolExecutor(
+            max_workers=1, thread_name_prefix="aip-wait-results"
+        ).submit(task.wait_for_results, self._gcs_fs, self._aip_client)
+
+    def _create_job_name(self, task_key):
+        # AIP job names must be alphanumeric (including underscore) and start
+        # with a letter.
+        alphanumeric_task_key = self.NON_ALPHANUMERIC_PATTERN.sub("_", str(task_key))
+        return f"bionic_{alphanumeric_task_key}_{uuid4().hex}"
 
 
 class ProcessExecutor:
