@@ -15,11 +15,9 @@ from uuid import uuid4
 from pathlib import Path
 
 from .datatypes import CodeFingerprint, Artifact
-from .gcs import get_gcs_fs_without_warnings
 from .utils.files import (
     ensure_dir_exists,
     ensure_parent_dir_exists,
-    recursively_copy_path,
 )
 from .utils.misc import hash_simple_obj_to_hex, oneline
 from .utils.urls import (
@@ -533,10 +531,7 @@ class Inventory:
                 tier=self.tier,
                 provenance=metadata_record.provenance,
                 exactly_matches_provenance=(match.level == "exact"),
-                artifact=Artifact(
-                    url=metadata_record.artifact_url,
-                    content_hash=metadata_record.value_hash,
-                ),
+                artifact=metadata_record.artifact,
             )
 
     def list_items(self):
@@ -548,7 +543,7 @@ class Inventory:
             metadata_record = self._load_metadata_if_valid_else_delete(metadata_url)
             if metadata_record is None:
                 continue
-            artifact_url = metadata_record.artifact_url
+            artifact_url = metadata_record.artifact.url
             yield ExternalCacheItem(
                 inventory=self,
                 abs_artifact_url=derelativize_url(artifact_url, metadata_url),
@@ -629,13 +624,13 @@ class Inventory:
         except Exception as e:
             raise InternalCacheStateError.from_failure("metadata record", url, e)
 
-        if not self._fs.exists(metadata_record.artifact_url):
+        if not self._fs.exists(metadata_record.artifact.url):
             logger.info(
                 "Found invalid metadata record at %s, "
                 "referring to nonexistent artifact at %s; "
                 "deleting metadata record",
                 url,
-                metadata_record.artifact_url,
+                metadata_record.artifact.url,
             )
             self.delete_url(url)
             return None
@@ -647,10 +642,9 @@ class Inventory:
         metadata_url = self._exact_metadata_url_for_provenance(provenance)
 
         metadata_record = ArtifactMetadataRecord(
+            artifact=artifact,
             descriptor=provenance.descriptor,
-            artifact_url=artifact.url,
             provenance=provenance,
-            value_hash=artifact.content_hash,
         )
         metadata_yaml = metadata_record.to_relativized_yaml(metadata_url)
         self._fs.write_bytes(metadata_yaml.encode("utf8"), metadata_url)
@@ -707,8 +701,8 @@ class GcsCloudStore:
     those blobs).
     """
 
-    def __init__(self, url):
-        self._fs = GcsFilesystem(url, "/inventory")
+    def __init__(self, gcs_fs, url):
+        self._fs = GcsFilesystem(gcs_fs, url, "/inventory")
 
         self.inventory = Inventory("GCS", "cloud", self._fs)
         self._artifact_root_url_prefix = url + "/artifacts"
@@ -754,32 +748,6 @@ class GcsCloudStore:
             self._fs.get_file(url, path)
 
 
-class FakeCloudStore(LocalStore):
-    """
-    A mock version of the GcsCloudStore that's actually backed by local files.
-    Useful for running tests without setting up a GCS connection, which is
-    slow and requires some configuration.
-    """
-
-    def __init__(self, root_path_str):
-        super(FakeCloudStore, self).__init__(root_path_str)
-
-    def generate_unique_url_prefix(self, provenance):
-        return url_from_path(self.generate_unique_dir_path(provenance))
-
-    def upload(self, path, url):
-        src_path = path
-        dst_path = path_from_url(url)
-
-        recursively_copy_path(src_path, dst_path)
-
-    def download(self, path, url):
-        src_path = path_from_url(url)
-        dst_path = path
-
-        recursively_copy_path(src_path, dst_path)
-
-
 class LocalFilesystem:
     """
     Implements a generic "FileSystem" interface for reading/writing small files
@@ -833,7 +801,7 @@ class GcsFilesystem:
     LocalFilesystem.
     """
 
-    def __init__(self, url, object_prefix_extension):
+    def __init__(self, gcs_fs, url, object_prefix_extension):
         if url.endswith("/"):
             url = url[:-1]
         self.root_url = url + object_prefix_extension
@@ -841,25 +809,7 @@ class GcsFilesystem:
         bucket_name, object_prefix = bucket_and_object_names_from_gs_url(url)
         self._bucket_name = bucket_name
         self._object_prefix = object_prefix
-        self._init_fs()
-
-    def __getstate__(self):
-        # Copy the object's state from self.__dict__ which contains
-        # all our instance attributes. Always use the dict.copy()
-        # method to avoid modifying the original state.
-        state = self.__dict__.copy()
-        # Remove the unpicklable entries.
-        del state["_fs"]
-        return state
-
-    def __setstate__(self, state):
-        # Restore instance attributes.
-        self.__dict__.update(state)
-        # Restore the fs.
-        self._init_fs()
-
-    def _init_fs(self):
-        self._fs = get_gcs_fs_without_warnings()
+        self._fs = gcs_fs
 
     def exists(self, url):
         self._validate_object_name_from_url(url)
@@ -876,8 +826,7 @@ class GcsFilesystem:
 
     def write_bytes(self, content_bytes, url):
         self._validate_object_name_from_url(url)
-        with self._fs.open(url, "wb") as f:
-            f.write(content_bytes)
+        self._fs.pipe(url, content_bytes)
 
     def read_bytes(self, url):
         self._validate_object_name_from_url(url)
@@ -948,9 +897,7 @@ def valid_filename_from_provenance(provenance):
     return provenance.descriptor.replace(" ", "-")
 
 
-# Next time we change this, it would be nice to also update ArtifactMetadataRecord to
-# contain an Artifact directly.
-CACHE_SCHEMA_VERSION = 10
+CACHE_SCHEMA_VERSION = 11
 
 
 class YamlRecordParsingError(Exception):
@@ -1158,17 +1105,16 @@ class ArtifactMetadataRecord:
     Describes a persisted artifact.  Intended to be stored as a YAML file.
     """
 
+    artifact: Artifact = attr.ib()
     descriptor: str = attr.ib()
-    artifact_url: str = attr.ib()
     provenance: Provenance = attr.ib()
-    value_hash: str = attr.ib()
 
     def to_relativized_yaml(self, metadata_url):
         """
         Serializes this object to a YAML string suitable for storage.
 
         The YAML representation directly maps to the structure of this class, except
-        that the ``artifact_url`` field will be converted to be relative to
+        that the ``artifact.url`` field will be converted to be relative to
         ``metadata_url``. This allows the artifact and metadata files to be moved
         around, as long as their relative position remains the same.
         """
@@ -1188,7 +1134,7 @@ class ArtifactMetadataRecord:
 
         This is the inverse operation to ``to_relativized_yaml``, and it correspondingly
         reverses the relativization of the artifact URL: the deserialized object will
-        have an absolute ``artifact_url`` field.
+        have an absolute ``artifact.url`` field.
         """
         try:
             body_dict = yaml.load(yaml_str, Loader=YamlLoader)
@@ -1206,11 +1152,15 @@ class ArtifactMetadataRecord:
     def _relativize(self, metadata_url):
         return attr.evolve(
             self,
-            artifact_url=relativize_url(self.artifact_url, metadata_url),
+            artifact=self.artifact.evolve(
+                url=relativize_url(self.artifact.url, metadata_url)
+            ),
         )
 
     def _derelativize(self, metadata_url):
         return attr.evolve(
             self,
-            artifact_url=derelativize_url(self.artifact_url, metadata_url),
+            artifact=self.artifact.evolve(
+                url=derelativize_url(self.artifact.url, metadata_url)
+            ),
         )
