@@ -23,11 +23,13 @@ import numpy as np
 from pyarrow import parquet, Table
 import pandas as pd
 
+from .code_hasher import CodeHasher
 from .decoration import decorator_updating_accumulator
 from .exception import EntityValueError, UnsupportedSerializedValueError
 from .deps.optdep import import_optional_dependency
 from .utils.files import recursively_copy_path
 from .utils.misc import (
+    hexdigest_from_path,
     oneline,
     read_hashable_bytes_from_file_or_dir,
     single_element,
@@ -47,6 +49,9 @@ def check_is_like_protocol(obj):
             )
 
 
+DEFAULT_VALUE_HASH_DIGEST = "DEFAULT_VALUE_HASH"
+
+
 class BaseProtocol:
     def validate(self, value):
         """
@@ -156,6 +161,13 @@ class BaseProtocol:
         """Like ``tokenize``, but operates on a serialized value."""
         return tokenization.tokenize(path, read_hashable_bytes_from_file_or_dir)
 
+    def get_extra_value_hash(self, value, suppress_warnings):
+        """
+        Generates additional data that can be added to the hash which doesn't
+        show up in the serialized output.
+        """
+        return DEFAULT_VALUE_HASH_DIGEST
+
     def _write_to_bytes(self, value):
         """
         Like ``write``, but returns an array of bytes rather than writing to a file.
@@ -273,6 +285,9 @@ class PicklableProtocol(BaseProtocol):
     def get_fixed_file_extension(self):
         return "pkl"
 
+    def get_extra_value_hash(self, value, suppress_warnings):
+        return CodeHasher.hash(type(value), suppress_warnings)
+
     def write(self, value, path):
         with path.open("wb") as file_:
             pickle.dump(value, file_, protocol=self._pickle_protocol_version)
@@ -293,6 +308,9 @@ class DillableProtocol(BaseProtocol):
     def get_fixed_file_extension(self):
         return "dill"
 
+    def get_extra_value_hash(self, value, suppress_warnings):
+        return CodeHasher.hash(type(value), suppress_warnings)
+
     def __init__(self, suppress_dill_side_effects=True):
         super(DillableProtocol, self).__init__()
 
@@ -599,6 +617,88 @@ class PathProtocol(BaseProtocol):
         return single_element(path.iterdir())
 
 
+# This protocol allows us to hash sets deterministically.
+# It would be preferable if PicklableProtocol could install a special handler for sets,
+# but unfortunately pickle's override mechanisms (dispatch tables and reduce_override)
+# don't work for built-in types like set -- at least not in the C implementation of
+# pickle. This custom protocol will work for set objects, but not for any other objects
+# that happen to contain sets.
+class PicklableSetProtocol(BaseProtocol):
+    """
+    Decorator indicating that an entity's value will be a set or frozenset whose
+    contents are picklable.
+
+    Sets need special handling because their iteration order is non-deterministic, so
+    the result of pickling them is also non-deterministic, which causes problems when
+    the serialized file is hashed. To resolve this, this protocol pickles each element
+    of a set into a separate file, which can then be hashed deterministically.
+
+    Parameters:
+        pickle_protocol_version: int (default: 4)
+            The pickle serialization protocol to use.
+    """
+
+    def __init__(self, pickle_protocol_version=4):
+        super(PicklableSetProtocol, self).__init__()
+        self._pickle_protocol_version = pickle_protocol_version
+
+    def get_fixed_file_extension(self):
+        return "setpkl"
+
+    def validate(self, value):
+        assert isinstance(value, (set, frozenset))
+
+    def write(self, value, path):
+        # We store each element of the set as a separate file, named after the hash of
+        # its contents; this guarantees that the whole directory will get a consistent
+        # hash value regardless of what order the elements are accessed in.
+        path.mkdir()
+
+        items_path = self.items_dir_sub_path(path)
+        items_path.mkdir()
+
+        occurrence_counts_by_hash = Counter()
+        for item in value:
+            item_file_path = items_path / "_tmp.pkl"
+            with item_file_path.open("wb") as file_:
+                pickle.dump(item, file_, protocol=self._pickle_protocol_version)
+
+            # We name each file based on a hash of its contents. However, it's possible
+            # for two objects in a set to generate the same hash (see
+            # `test_set_with_duplicate_pickle_output` for an example), so we add an
+            # extra number that increments each time we see the same hash.
+            item_file_hash = hexdigest_from_path(item_file_path)
+            n_identical_files_so_far = occurrence_counts_by_hash[item_file_hash]
+            occurrence_counts_by_hash[item_file_hash] += 1
+
+            filename = f"{item_file_hash}_{n_identical_files_so_far}.pkl"
+            new_item_file_path = items_path / filename
+            assert not new_item_file_path.exists()
+            item_file_path.rename(new_item_file_path)
+
+        type_path = self.type_file_sub_path(path)
+        type_path.write_bytes(
+            pickle.dumps(type(value), protocol=self._pickle_protocol_version)
+        )
+
+    def read(self, path):
+        type_path = self.type_file_sub_path(path)
+        collection_type = pickle.loads(type_path.read_bytes())
+
+        items = []
+        for item_file_path in self.items_dir_sub_path(path).iterdir():
+            with item_file_path.open("rb") as file_:
+                items.append(pickle.load(file_))
+
+        return collection_type(items)
+
+    def type_file_sub_path(self, path):
+        return path / "type.pkl"
+
+    def items_dir_sub_path(self, path):
+        return path / "items"
+
+
 class CombinedProtocol(BaseProtocol):
     """
     Decorator generator indicating that an entity's values should be handled
