@@ -17,7 +17,10 @@ def pytest_addoption(parser):
         "--bucket", action="store", help="URL to GCS bucket to use for tests"
     )
     parser.addoption(
-        "--project", action="store", help="GCP project to use for AIP tests"
+        "--aip",
+        action="store_true",
+        default=False,
+        help="run AIP tests, requires --bucket",
     )
     parser.addoption(
         "--parallel",
@@ -41,6 +44,8 @@ def pytest_configure(config):
         "allows_parallel",
         "can run with parallel execution even when that's not explicitly enabled",
     )
+    add_mark("real_gcp_only", "runs on real GCP only")
+    add_mark("fake_gcp_only", "runs on fake GCP only")
 
     # These markers are added automatically based on parametric fixtures.
     add_mark("serial", "will run using serial execution")
@@ -59,9 +64,9 @@ def pytest_collection_modifyitems(config, items):
     has_gcs = config.getoption("--bucket")
     skip_needs_gcs = pytest.mark.skip(reason="only runs when --bucket is set")
 
-    has_aip = config.getoption("--project") and has_gcs
+    has_aip = has_gcs and config.getoption("--aip")
     skip_needs_aip = pytest.mark.skip(
-        reason="only runs when --bucket and --project are set"
+        reason="only runs when both --bucket and --aip are set"
     )
 
     also_run_parallel = config.getoption("--parallel")
@@ -76,6 +81,9 @@ def pytest_collection_modifyitems(config, items):
                 item.add_marker(skip_slow)
 
         if "real_gcp" in item.keywords:
+            if "fake_gcp_only" in item.keywords:
+                continue
+
             if "needs_gcs" in item.keywords:
                 item_is_baseline = False
                 if not has_gcs:
@@ -86,6 +94,10 @@ def pytest_collection_modifyitems(config, items):
                 if not has_aip:
                     item.add_marker(skip_needs_aip)
 
+        elif "fake_gcp" in item.keywords:
+            if "real_gcp_only" in item.keywords:
+                continue
+
         if "parallel" in item.keywords:
             if "allows_parallel" not in item.keywords:
                 item_is_baseline = False
@@ -93,9 +105,6 @@ def pytest_collection_modifyitems(config, items):
                 if "no_parallel" in item.keywords or not also_run_parallel:
                     continue
 
-            if "fake_gcp" in item.keywords:
-                continue
-
         elif "needs_parallel" in item.keywords:
             continue
 
