@@ -1,73 +1,21 @@
 import io
-import logging
-from contextlib import contextmanager
-from functools import partial
-from pathlib import Path
 import re
+from concurrent.futures import Future
+from concurrent.futures.process import ProcessPoolExecutor
+from contextlib import contextmanager
+from pathlib import Path
+from tempfile import mkdtemp
+from typing import Dict
 from unittest.mock import Mock
-from uuid import uuid4
 
-from bionic import gcs
-from bionic.aip import client as aip_client
-from bionic.aip.future import Future as AipFuture
+import cloudpickle
+
 from bionic.aip.main import _run as run_aip
-from bionic.aip.task import Task as AipTask
+from bionic.persistence import LocalStore
 from bionic.utils.files import ensure_parent_dir_exists
 from bionic.utils.urls import path_from_url
 
 
-class FakeAipFuture(AipFuture):
-    """
-    A mock implementation of Future that finished successfully.
-    """
-
-    def __init__(self, project_name: str, job_id: str, output: str):
-        self.project_name = project_name
-        self.job_id = job_id
-        self.output = output
-
-    def _get_state_and_error(self):
-        from bionic.aip.future import State
-
-        return State.SUCCEEDED, ""
-
-
-class FakeAipTask(AipTask):
-    """
-    A mock implementation of Task that runs the job locally using a
-    subprocess instead of running it on AIP.
-    """
-
-    def submit(self) -> AipFuture:
-        self._stage()
-        spec = self._ai_platform_job_spec()
-
-        logging.info(f"Submitting test {self.config.project_name}: {self}")
-        import subprocess
-
-        subprocess.check_call(spec["trainingInput"]["args"])
-        return FakeAipFuture(self.config.project_name, self.job_id(), self.output_uri())
-
-
-class FakeAipExecutor:
-    """
-    A mock version of AipExecutor to submit FakeTasks that uses
-    subprocess to execute the tasks instead of using AIP. Useful for
-    running tests locally without using AIP.
-    """
-
-    def __init__(self, aip_config):
-        self._aip_config = aip_config
-
-    def submit(self, task_config, fn, *args, **kwargs):
-        return FakeAipTask(
-            name="a" + str(uuid4()).replace("-", ""),
-            config=self._aip_config,
-            task_config=task_config,
-            function=partial(fn, *args, **kwargs),
-        ).submit()
-
-
 # TODO It would be nice to use a different fsspec implementation here instead
 # of writing a custom one. We tried fsspec.implementations.memory.MemoryFileSystem
 # implementation, but that has a few problems.
@@ -80,8 +28,8 @@ class FakeAipExecutor:
 # We can override 1) but we need to get 2) fixed in fsspec before we can replace
 # this implementation.
 class FakeGcsFs:
-    def __init__(self):
-        self._files_by_url = {}
+    def __init__(self, shared_dict):
+        self._files_by_url: dict = shared_dict
 
     def cat_file(self, url):
         with self.open(url, "rb") as f:
@@ -118,7 +66,7 @@ class FakeGcsFs:
         url = url[:-4]
         # Glob doesn't return the protocol, so we strip "gs://".
         urls = []
-        for file_url in self._files_by_url:
+        for file_url in self._files_by_url.keys():
             assert file_url.startswith("gs://")
             if url in file_url:
                 urls.append(file_url[5:])
@@ -127,10 +75,12 @@ class FakeGcsFs:
     def isdir(self, url):
         if not url.endswith("/"):
             url = url + "/"
-        return any(key for key in self._files_by_url if key.startswith(url))
+        return any(key for key in self._files_by_url.keys() if key.startswith(url))
 
     @contextmanager
     def open(self, url, mode):
+        assert url.startswith("gs://")
+
         assert mode in ("rb", "wb")
         if mode == "rb" and url not in self._files_by_url:
             raise FileNotFoundError(f"no file found for url {url}")
@@ -166,37 +116,77 @@ class FakeGcsFs:
         else:
             del self._files_by_url[url]
 
+    def pipe(self, url, content_bytes):
+        with self.open(url, "wb") as f:
+            f.write(content_bytes)
 
-@contextmanager
-def run_in_fake_gcp(fake_gcs_fs: FakeGcsFs, caplog):
-    """
-    Use fake GCP by mocking out GCS and AIP.
-    """
 
-    def create_aip_job(body, parent):
-        # AIP executions do not transmit logs to local instance of bionic.
-        # Emulate this behavior by setting the log level; this should filter out
-        # all (or almost all) the logs. Tests that check the log output will
-        # only test against logs which are printed locally.
-        with caplog.at_level(logging.CRITICAL):
-            run_aip(body["trainingInput"]["args"][3])
+class FakeAipClient:
+    def __init__(self, gcs_fs, tmp_path):
+        self._gcs_fs = gcs_fs
+        self._tmp_path = tmp_path
+        self._jobs: Dict[str, Future] = {}
+
+    def _create_aip_job(self, body, parent):
+        assert body["trainingInput"]["args"][:3] == [
+            "python",
+            "-m",
+            "bionic.aip.main",
+        ]
+
+        input_uri = body["trainingInput"]["args"][3]
+
+        # AIP execution does not have access to the disk cache of the local
+        # bionic instance. In order to emulate disk filesystem isolation, we can
+        # change location of disk cache for the AIP job.
+        self._modify_disk_cache_location(input_uri)
+
+        self._jobs[body["jobId"]] = ProcessPoolExecutor(max_workers=1).submit(
+            run_aip, input_uri, self._gcs_fs
+        )
+
         return Mock()
 
-    mock_aip_client = Mock()
-    mock_aip_client.projects().jobs().create = create_aip_job
-    mock_aip_client.projects().jobs().get().execute.return_value = {
-        "state": "SUCCEEDED"
-    }
+    def _modify_disk_cache_location(self, input_uri):
+        assert self._gcs_fs.exists(input_uri)
+        with self._gcs_fs.open(input_uri, "rb") as f:
+            task = cloudpickle.load(f)
 
-    cached_gcs_fs = gcs._cached_gcs_fs
-    cached_aip_client = aip_client._cached_aip_client
-    try:
-        gcs._cached_gcs_fs = fake_gcs_fs
-        aip_client._cached_aip_client = mock_aip_client
-        yield
-    finally:
-        gcs._cached_gcs_fs = cached_gcs_fs
-        aip_client._cached_aip_client = cached_aip_client
+        task.function.args[0]._context.core.persistent_cache._local_store = LocalStore(
+            mkdtemp(dir=self._tmp_path)
+        )
+
+        with self._gcs_fs.open(input_uri, "wb") as f:
+            cloudpickle.dump(task, f)
+
+    def _get_aip_job(self, name):
+        job_id = name.split("/")[-1]
+
+        assert job_id in self._jobs
+        future = self._jobs[job_id]
+
+        if future.done():
+            try:
+                future.result()
+                response = {"state": "SUCCEEDED"}
+            except Exception as e:
+                # Exception information is only added for convenience. In real
+                # GCP, if an AIP job fails, the error message will contain a url
+                # instead.
+                response = {"state": "FAILED", "errorMessage": repr(e)}
+        else:
+            response = {"state": "RUNNING"}
+
+        job = Mock()
+        job.execute.return_value = response
+        return job
+
+    def projects(self):
+        # TODO: check that aip_task_config tasks were actually sent to AIP.
+        projects = Mock()
+        projects.jobs().create = self._create_aip_job
+        projects.jobs().get = self._get_aip_job
+        return projects
 
 
 class InstrumentedFilesystem:
@@ -207,26 +197,20 @@ class InstrumentedFilesystem:
 
     def __init__(self, wrapped_fs, make_list):
         self._wrapped_fs = wrapped_fs
-        self._downloaded_urls = make_list()
-        self._uploaded_urls = make_list()
+        self._downloaded_urls: list = make_list()
+        self._uploaded_urls: list = make_list()
 
     def get_file(self, url, path):
         self._downloaded_urls.append(url)
         return self._wrapped_fs.get_file(url, path)
 
-    def get_dir(self, url, path):
-        self._downloaded_urls.append(url)
-        return self._wrapped_fs.get_dir(url, path)
-
     def put_file(self, path, url):
         self._uploaded_urls.append(url)
         return self._wrapped_fs.put_file(path, url)
 
-    def put_dir(self, path, url):
-        self._uploaded_urls.append(url)
-        return self._wrapped_fs.put_dir(path, url)
-
     def __getattr__(self, attr):
+        if "_wrapped_fs" not in vars(self):
+            raise AttributeError
         return getattr(self._wrapped_fs, attr)
 
     def matching_urls_downloaded(self, url_regex):
@@ -249,22 +233,3 @@ class InstrumentedFilesystem:
         match_count = len(items) - len(non_matching_items)
         items[:] = non_matching_items
         return match_count
-
-
-# TODO It would be nice if the GCS filesystem (and indeed all filesystems) were
-# injected rather than accessed (and patched) globally.
-@contextmanager
-def instrument_gcs_fs(make_list):
-    """
-    Context manager that temporarily globally patches the
-    ``gcs.get_gcs_fs_without_warnings`` function, replacing the usual filesystem with
-    an instrumented version that tracks all upload and download operations.
-    """
-    orig_gcs_fs = gcs._cached_gcs_fs
-    inst_gcs_fs = InstrumentedFilesystem(orig_gcs_fs, make_list)
-    gcs._cached_gcs_fs = inst_gcs_fs
-
-    try:
-        yield inst_gcs_fs
-    finally:
-        gcs._cached_gcs_fs = orig_gcs_fs
