@@ -20,7 +20,7 @@ from bionic.exception import (
     UnsetEntityError,
 )
 
-from ..helpers import assert_re_matches, count_calls
+from ..helpers import assert_re_matches
 
 
 @pytest.fixture(scope="function")
@@ -606,7 +606,7 @@ def test_in_memory_caching(builder, make_counter):
 
     @builder
     @bn.persist(False)
-    @count_calls(counter)
+    @counter
     def xy(x, y):
         return x * y
 
@@ -738,7 +738,10 @@ def test_multiple_compute_attempts(builder):
 # test after finding a bug where Bionic was applying == to a fixed value, which doesn't
 # work for DataFrames.)
 def test_fixed_dataframe(builder):
-    df = pd.DataFrame(columns=["a", "b"], data=[[1, 2], [3, 4]],)
+    df = pd.DataFrame(
+        columns=["a", "b"],
+        data=[[1, 2], [3, 4]],
+    )
 
     builder.assign("df", df)
 
