@@ -1,9 +1,10 @@
 import pytest
 
 import math
+from textwrap import dedent
 import threading
 
-from ..helpers import RoundingProtocol
+from ..helpers import RoundingProtocol, import_code
 from bionic.exception import AttributeValidationError, CodeVersioningError
 from bionic.protocols import PicklableProtocol
 
@@ -245,7 +246,7 @@ def test_versioning(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x * y
@@ -256,7 +257,7 @@ def test_versioning(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return y * x
@@ -267,7 +268,7 @@ def test_versioning(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(major=1, minor=1)
+    @bn.version_no_warnings(major=1, minor=1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return y * x
@@ -276,7 +277,7 @@ def test_versioning(builder, make_counter):
     assert call_counter.times_called() == 0
 
     @builder  # noqa: F811
-    @bn.version(major=1, minor=1)
+    @bn.version_no_warnings(major=1, minor=1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x ** y
@@ -287,7 +288,7 @@ def test_versioning(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(major=2)
+    @bn.version_no_warnings(major=2)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x ** y
@@ -326,7 +327,7 @@ def test_indirect_versioning(builder, make_counter):
     assert f_call_counter.times_called() == 0
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return 4
@@ -336,7 +337,7 @@ def test_indirect_versioning(builder, make_counter):
     assert f_call_counter.times_called() == 1
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return len("xxxx")
@@ -346,7 +347,7 @@ def test_indirect_versioning(builder, make_counter):
     assert f_call_counter.times_called() == 0
 
     @builder  # noqa: F811
-    @bn.version(1, minor=1)
+    @bn.version_no_warnings(1, minor=1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return len("xxxx")
@@ -398,7 +399,7 @@ def test_versioning_assist(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x * y
@@ -409,7 +410,7 @@ def test_versioning_assist(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return y * x
@@ -420,7 +421,7 @@ def test_versioning_assist(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(major=1, minor=1)
+    @bn.version_no_warnings(major=1, minor=1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return y * x
@@ -429,7 +430,7 @@ def test_versioning_assist(builder, make_counter):
     assert call_counter.times_called() == 0
 
     @builder  # noqa: F811
-    @bn.version(major=1, minor=1)
+    @bn.version_no_warnings(major=1, minor=1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x ** y
@@ -440,7 +441,7 @@ def test_versioning_assist(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(major=2)
+    @bn.version_no_warnings(major=2)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x ** y
@@ -449,6 +450,72 @@ def test_versioning_assist(builder, make_counter):
     assert call_counter.times_called() == 1
 
 
+def test_versioning_assist_with_refs(builder, make_counter):
+    call_counter = make_counter()
+
+    builder.set("core__versioning_mode", "assist")
+
+    builder.assign("x", 2)
+    builder.assign("y", 3)
+
+    def op(x, y):
+        return x + y
+
+    @builder
+    def f(x, y):
+        call_counter.mark()
+        return op(x, y)
+
+    assert builder.build().get("f") == 5
+    assert builder.build().get("f") == 5
+    assert call_counter.times_called() == 1
+
+    def op(x, y):  # noqa: F811
+        return x * y
+
+    with raises_versioning_error_for_entity("f"):
+        builder.build().get("f")
+
+    @builder  # noqa: F811
+    @bn.version_no_warnings(1)
+    def f(x, y):  # noqa: F811
+        call_counter.mark()
+        return op(x, y)
+
+    assert builder.build().get("f") == 6
+    assert call_counter.times_called() == 1
+
+    def op(x, y):  # noqa: F811
+        return y * x
+
+    with raises_versioning_error_for_entity("f"):
+        builder.build().get("f")
+
+    @builder  # noqa: F811
+    @bn.version_no_warnings(major=1, minor=1)
+    def f(x, y):  # noqa: F811
+        call_counter.mark()
+        return op(x, y)
+
+    assert builder.build().get("f") == 6
+    assert call_counter.times_called() == 0
+
+    def op(x, y):  # noqa: F811
+        return x ** y
+
+    with raises_versioning_error_for_entity("f"):
+        builder.build().get("f")
+
+    @builder  # noqa: F811
+    @bn.version_no_warnings(major=2)
+    def f(x, y):  # noqa: F811
+        call_counter.mark()
+        return op(x, y)
+
+    assert builder.build().get("f") == 8
+    assert call_counter.times_called() == 1
+
+
 def test_indirect_versioning_assist(builder, make_counter):
     y_call_counter = make_counter()
     f_call_counter = make_counter()
@@ -480,7 +547,7 @@ def test_indirect_versioning_assist(builder, make_counter):
         builder.build().get("f")
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return 4
@@ -490,7 +557,7 @@ def test_indirect_versioning_assist(builder, make_counter):
     assert f_call_counter.times_called() == 1
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return len("xxxx")
@@ -499,7 +566,7 @@ def test_indirect_versioning_assist(builder, make_counter):
         builder.build().get("f")
 
     @builder  # noqa: F811
-    @bn.version(1, minor=1)
+    @bn.version_no_warnings(1, minor=1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return len("xxxx")
@@ -618,7 +685,7 @@ def test_versioning_auto(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x * y
@@ -629,7 +696,7 @@ def test_versioning_auto(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return y * x
@@ -640,7 +707,7 @@ def test_versioning_auto(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(major=1, minor=1)
+    @bn.version_no_warnings(major=1, minor=1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return y * x
@@ -649,7 +716,7 @@ def test_versioning_auto(builder, make_counter):
     assert call_counter.times_called() == 0
 
     @builder  # noqa: F811
-    @bn.version(major=1, minor=1)
+    @bn.version_no_warnings(major=1, minor=1)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x ** y
@@ -660,7 +727,7 @@ def test_versioning_auto(builder, make_counter):
     builder.delete("f")
 
     @builder  # noqa: F811
-    @bn.version(major=2)
+    @bn.version_no_warnings(major=2)
     def f(x, y):  # noqa: F811
         call_counter.mark()
         return x ** y
@@ -701,7 +768,7 @@ def test_indirect_versioning_auto(builder, make_counter):
     assert f_call_counter.times_called() == 1
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return 4
@@ -713,7 +780,7 @@ def test_indirect_versioning_auto(builder, make_counter):
     assert f_call_counter.times_called() == 0
 
     @builder  # noqa: F811
-    @bn.version(1)
+    @bn.version_no_warnings(1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return len("xxxx")
@@ -725,7 +792,7 @@ def test_indirect_versioning_auto(builder, make_counter):
     assert f_call_counter.times_called() == 0
 
     @builder  # noqa: F811
-    @bn.version(1, minor=1)
+    @bn.version_no_warnings(1, minor=1)
     def y():  # noqa: F811
         y_call_counter.mark()
         return len("xxxx")
@@ -747,6 +814,312 @@ def test_indirect_versioning_auto(builder, make_counter):
     assert f_call_counter.times_called() == 0
 
 
+def test_versioning_auto_class_changes(builder, make_counter):
+    call_counter = make_counter()
+
+    # PicklableProtocol cannot pickle IntWrapper class because it is
+    # defined inside the test function. We instead use this custom
+    # protocol that wraps the PicklableProtocol to be able to serialize
+    # IntWrapper.
+    class IntWrapperProtocol(bn.protocols.PicklableProtocol):
+        def write(self, wrapper, path):
+            assert isinstance(wrapper, IntWrapper)
+            super().write(wrapper.value, path)
+
+        def read(self, path):
+            return IntWrapper(super().read(path))
+
+    class IntWrapper:
+        def __init__(self, value):
+            self.value = value
+
+        def __str__(self):
+            return f"IntWrapper: value={self.value}"
+
+    builder.set("core__versioning_mode", "auto")
+
+    builder.assign("x", 2)
+    builder.assign("y", 3)
+
+    @builder
+    @IntWrapperProtocol()
+    def wrapped_f(x, y):
+        return IntWrapper(x + y)
+
+    @builder
+    def f(wrapped_f):
+        call_counter.mark()
+        return wrapped_f.value
+
+    assert builder.build().get("f") == 5
+    assert builder.build().get("f") == 5
+    assert call_counter.times_called() == 1
+
+    builder.delete("wrapped_f")
+
+    @builder  # noqa: F811
+    @IntWrapperProtocol()
+    def wrapped_f(x, y):  # noqa: F811
+        return IntWrapper(x * y)
+
+    assert builder.build().get("f") == 6
+    assert call_counter.times_called() == 1
+
+    builder.delete("wrapped_f")
+
+    # No changes in the wrapper class.
+    class IntWrapper:  # noqa: F811
+        def __init__(self, value):
+            self.value = value
+
+        def __str__(self):
+            return f"IntWrapper: value={self.value}"
+
+    @builder  # noqa: F811
+    @IntWrapperProtocol()
+    def wrapped_f(x, y):  # noqa: F811
+        return IntWrapper(x * y)
+
+    assert builder.build().get("f") == 6
+    assert call_counter.times_called() == 0
+
+    builder.delete("wrapped_f")
+
+    # Make a change in the __str__ method.
+    class IntWrapper:  # noqa: F811
+        def __init__(self, value):
+            self.value = value
+
+        def __str__(self):
+            return f"value={self.value}"
+
+    @builder  # noqa: F811
+    @IntWrapperProtocol()
+    def wrapped_f(x, y):  # noqa: F811
+        return IntWrapper(x * y)
+
+    assert builder.build().get("f") == 6
+    assert call_counter.times_called() == 1
+
+
+@pytest.mark.parametrize("versioning_mode", ["manual", "assist"])
+def test_versioning_class_changes(builder, make_counter, versioning_mode):
+    call_counter = make_counter()
+
+    # PicklableProtocol cannot pickle IntWrapper class because it is
+    # defined inside the test function. We instead use this custom
+    # protocol that wraps the PicklableProtocol to be able to serialize
+    # IntWrapper.
+    class IntWrapperProtocol(bn.protocols.PicklableProtocol):
+        def write(self, wrapper, path):
+            assert isinstance(wrapper, IntWrapper)
+            super().write(wrapper.value, path)
+
+        def read(self, path):
+            return IntWrapper(super().read(path))
+
+    class IntWrapper:
+        def __init__(self, value):
+            self.value = value
+
+        def __str__(self):
+            return f"IntWrapper: value={self.value}"
+
+    builder.set("core__versioning_mode", versioning_mode)
+
+    builder.assign("x", 2)
+    builder.assign("y", 3)
+
+    @builder
+    @IntWrapperProtocol()
+    @bn.version(1)
+    def wrapped_f(x, y):
+        return IntWrapper(x + y)
+
+    @builder
+    def f(wrapped_f):
+        call_counter.mark()
+        return wrapped_f.value
+
+    assert builder.build().get("f") == 5
+    assert builder.build().get("f") == 5
+    assert call_counter.times_called() == 1
+
+    builder.delete("wrapped_f")
+
+    # Make a change in the __str__ method.
+    class IntWrapper:  # noqa: F811
+        def __init__(self, value):
+            self.value = value
+
+        def __str__(self):
+            return f"value={self.value}"
+
+    @builder  # noqa: F811
+    @IntWrapperProtocol()
+    @bn.version(1)
+    def wrapped_f(x, y):  # noqa: F811
+        return IntWrapper(x + y)
+
+    if versioning_mode == "assist":
+        with raises_versioning_error_for_entity("wrapped_f"):
+            builder.build().get("f")
+    else:
+        assert builder.build().get("f") == 5
+    assert call_counter.times_called() == 0
+
+    @builder  # noqa: F811
+    @IntWrapperProtocol()
+    @bn.version(2)
+    def wrapped_f(x, y):  # noqa: F811
+        return IntWrapper(x + y)
+
+    assert builder.build().get("f") == 5
+    # We don't hash the class of the inputs for manual and assist
+    # versioning modes. So even if the class for `wrapped_f` changes,
+    # if the hash of the `wrapped_f.value` doesn't changes, `f` won't
+    # be recomputed.
+    assert call_counter.times_called() == 0
+
+
+def test_versioning_auto_with_local_refs(builder, make_counter):
+    call_counter = make_counter()
+
+    builder.set("core__versioning_mode", "auto")
+
+    ret_val = 1
+
+    def ref_func():
+        return ret_val
+
+    @builder
+    @call_counter
+    def x():
+        return ref_func()
+
+    assert builder.build().get("x") == 1
+    assert call_counter.times_called() == 1
+
+    # With no changes, x is not called again.
+    assert builder.build().get("x") == 1
+    assert call_counter.times_called() == 0
+
+    # If we change the ref_func, the bytecode of x changes.
+    def ref_func():  # noqa: F811
+        return ret_val + 7
+
+    assert builder.build().get("x") == 8
+    assert call_counter.times_called() == 1
+
+    # If we change the ret_val, the bytecode of x changes.
+    ret_val = 5
+    assert builder.build().get("x") == 12
+    assert call_counter.times_called() == 1
+
+
+def test_versioning_auto_with_module_refs(builder, make_counter):
+    call_counter = make_counter()
+
+    builder.set("core__versioning_mode", "auto")
+
+    mod_code = """
+    ret_val = 1
+
+    def ref_func():
+        return ret_val
+    """
+    m = import_code(dedent(mod_code))
+
+    @builder
+    @call_counter
+    def x():
+        return m.ref_func()
+
+    assert builder.build().get("x") == 1
+    assert call_counter.times_called() == 1
+
+    # With no changes, x is not called again.
+    m = import_code(dedent(mod_code))
+    assert builder.build().get("x") == 1
+    assert call_counter.times_called() == 0
+
+    # If we change the ref_func, the bytecode of x changes.
+    mod_code = """
+    ret_val = 1
+
+    def ref_func():
+        return ret_val + 7
+    """
+    m = import_code(dedent(mod_code))
+    assert builder.build().get("x") == 8
+    assert call_counter.times_called() == 1
+
+    # If we change the ret_val, the bytecode of x changes.
+    mod_code = """
+    ret_val = 5
+
+    def ref_func():
+        return ret_val + 7
+    """
+    m = import_code(dedent(mod_code))
+    assert builder.build().get("x") == 12
+    assert call_counter.times_called() == 1
+
+
+def test_versioning_auto_version_options(builder):
+    builder.set("core__versioning_mode", "auto")
+
+    import pandas as pd
+
+    series = pd.Series([1, 2, 3])
+
+    @builder
+    def x():
+        assert series is not None
+        return 1
+
+    with pytest.warns(None) as warnings:
+        assert builder.build().get("x") == 1
+    assert len(warnings) == 0
+
+    # With suppress_bytecode_warnings as False, bytecode analysis
+    # throws a warning.
+    @builder  # noqa: F811
+    @bn.version(major=2)
+    def x():  # noqa: F811
+        assert series is not None
+        return 1
+
+    with pytest.warns(UserWarning, match="Found a complex object"):
+        assert builder.build().get("x") == 1
+
+    # With ignore_bytecode=True, we won't analyze bytecode and won't
+    # throw a warning.
+    @builder  # noqa: F811
+    @bn.version(major=2, ignore_bytecode=True)
+    def x():  # noqa: F811
+        assert series is not None
+        return 2
+
+    with pytest.warns(None) as warnings:
+        # It's a little awkward that we compute the value again even
+        # though we ignore the bytecode. But that's because the
+        # bytecode is now null which changes the version.
+        assert builder.build().get("x") == 2
+    assert len(warnings) == 0
+
+    @builder  # noqa: F811
+    @bn.version(major=2, ignore_bytecode=True)
+    def x():  # noqa: F811
+        assert series is not None
+        return 3
+
+    with pytest.warns(None) as warnings:
+        # Even with the change in bytecode, it uses the cached value.
+        assert builder.build().get("x") == 2
+    assert len(warnings) == 0
+
+
 def test_all_returned_results_are_deserialized(builder, make_counter):
     counter = make_counter()
 
@@ -1361,6 +1734,11 @@ def make_tracked_class():
     """
     Creates a "tracked" class which keeps a count of how many of its instances are in
     memory.
+
+    The class also asserts that each instance in memory is has a unique value; this
+    guarantees that we don't (for example) read an instance from disk while the
+    original instance is still in memory.
+
     """
 
     def _make_tracked_class():
@@ -1378,14 +1756,22 @@ def make_tracked_class():
 
         class Tracked:
             n_instances_in_memory = 0
+            instance_values_in_memory = set()
             protocol = TrackedProtocol()
 
             def __init__(self, value):
+                assert value not in self.__class__.instance_values_in_memory
+
+                self.__class__.instance_values_in_memory.add(value)
                 self.__class__.n_instances_in_memory += 1
                 self.value = value
 
             def __del__(self):
-                self.__class__.n_instances_in_memory -= 1
+                # If our assertion in __init__ failed, we don't have a "value"
+                # attribute and don't need to update our trackers.
+                if hasattr(self, "value"):
+                    self.__class__.n_instances_in_memory -= 1
+                    self.__class__.instance_values_in_memory.remove(self.value)
 
         return Tracked
 
@@ -1458,3 +1844,49 @@ def test_non_memoized_complex_tuples_are_garbage_collected(builder, make_tracked
         return x_plus_one + 1
 
     assert builder.build().get("x_plus_two") == 3
+
+
+# When an entity has both persistence and memoization disabled, we memoize it just for
+# the duration of the get() call. However, we may want to disable this behavior in the
+# future; for the time being it's configurable (but undocumented).
+@pytest.mark.parametrize("disable_query_caching", [False, True])
+def test_uncached_values_cached_for_query_duration(
+    builder, make_counter, disable_query_caching
+):
+    x_counter = make_counter()
+
+    if disable_query_caching:
+        builder.set("core__temp_memoize_if_uncached", False)
+        expected_x_calls_per_query = 2
+    else:
+        expected_x_calls_per_query = 1
+
+    builder.set("core__persist_by_default", False)
+    builder.set("core__memoize_by_default", False)
+
+    builder.assign("y", 3)
+
+    @builder
+    @x_counter
+    def x():
+        return 2
+
+    @builder
+    def x_squared(x):
+        return x * x
+
+    @builder
+    def y_x_squared(y, x_squared):
+        return y * x_squared
+
+    @builder
+    def y_x_cubed(y_x_squared, x):
+        return y_x_squared * x
+
+    flow = builder.build()
+
+    assert flow.get("y_x_cubed") == 24
+    assert x_counter.times_called() == expected_x_calls_per_query
+
+    assert flow.get("y_x_cubed") == 24
+    assert x_counter.times_called() == expected_x_calls_per_query
