@@ -10,6 +10,7 @@ from bionic.descriptors.parsing import dnode_from_descriptor
 from bionic.exception import CodeVersioningError
 from bionic.utils.misc import single_element, single_unique_element
 import bionic as bn
+from ..helpers import import_code
 
 
 class ModelFlowHarness:
@@ -28,6 +29,7 @@ class ModelFlowHarness:
     - testing the value of an entity and the set of (user-provided) functions computed
       in the process
     - persisted and non-persisted entities
+    - memoized and non-memoized entities
     - functions with multiple entity outputs
 
     This class is similar to the ``SimpleFlowModel`` ("SFM") in ``test_persisted_fuzz``,
@@ -37,7 +39,6 @@ class ModelFlowHarness:
     - this class does not support the following:
       - non-deterministic (``changes_per_run``) entities
       - non-functional updates
-      - alternative versioning modes
       - cloud caching
 
     This class has a more sophisticated implementation than SFM, so it should be
@@ -45,35 +46,61 @@ class ModelFlowHarness:
     the above features, we can remove SFM altogether.
     """
 
-    def __init__(self, builder, make_list):
+    def __init__(
+        self,
+        builder,
+        make_list,
+        parallel_execution_enabled=False,
+    ):
         self._builder = builder
         self._versioning_mode = "manual"
+        self._parallel_execution_enabled = parallel_execution_enabled
+        self._expect_exact_call_count_matches = True
+        self._query_caching_enabled = True
 
+        # TODO Would it make sense to factor all the model state into a single ModelFlow
+        # object?
         self._entities_by_name = {}
 
         self._descriptors_computed_by_flow = make_list()
         self._descriptors_computed_by_model = make_list()
 
+        self._clear_active_flow()
+
     def set_versioning_mode(self, mode):
         assert mode in ("manual", "assist", "auto")
         self._versioning_mode = mode
         self._builder.set("core__versioning_mode", mode)
 
+    def disable_query_caching(self):
+        self._builder.set("core__temp_memoize_if_uncached", False)
+        self._query_caching_enabled = False
+
+    def disable_exact_call_counting(self):
+        self._expect_exact_call_count_matches = False
+
     def get_all_entity_names(self):
         return list(self._entities_by_name.keys())
 
-    def create_entity(self, should_persist=True):
+    def create_entity(self, should_persist=True, should_memoize=True):
         name = f"e{len(self._entities_by_name) + 1}"
-        entity = ModelEntity(name=name, should_persist=should_persist)
+        entity = ModelEntity(
+            name=name,
+            should_persist=should_persist,
+            should_memoize=should_memoize,
+        )
         assert name not in self._entities_by_name
         self._entities_by_name[name] = entity
         return name
 
     def add_binding(self, out_descriptor, dep_entity_names, use_tuples_for_output=True):
+        self._clear_active_flow()
+
         out_dnode = dnode_from_descriptor(out_descriptor)
         out_entities = list(
             map(self._entities_by_name.get, out_dnode.all_entity_names())
         )
+        dep_entities = list(map(self._entities_by_name.get, dep_entity_names))
 
         should_persists = set(entity.should_persist for entity in out_entities)
         (should_persist,) = should_persists
@@ -81,7 +108,7 @@ class ModelFlowHarness:
         binding = ModelBinding(
             out_dnode=out_dnode,
             out_entities=out_entities,
-            dep_entity_names=dep_entity_names,
+            dep_entities=dep_entities,
             use_tuples_for_output=use_tuples_for_output,
         )
         for out_entity in out_entities:
@@ -99,6 +126,8 @@ class ModelFlowHarness:
         change_func_version=True,
         update_version_annotation=True,
     ):
+        self._clear_active_flow()
+
         entity = self._entities_by_name[entity_name]
         entity.binding.update(
             change_func=change_func_version,
@@ -108,18 +137,29 @@ class ModelFlowHarness:
 
         self._add_binding_to_flow(entity.binding)
 
-    def query_and_check_entity(self, entity_name):
-        flow = self._builder.build()
+    def query_and_check_entity(self, entity_name, expect_exact_call_count_matches=None):
+        if expect_exact_call_count_matches is None:
+            expect_exact_call_count_matches = self._expect_exact_call_count_matches
+
+        self._build_active_flow_if_missing()
 
         flow_exception = None
         try:
-            flow_value = flow.get(entity_name)
+            flow_value = self._active_flow.get(entity_name)
         except CodeVersioningError as e:
             flow_exception = e
 
+        context = ModelExecutionContext(
+            parallel_execution_enabled=self._parallel_execution_enabled,
+            query_caching_enabled=self._query_caching_enabled,
+            versioning_mode=self._versioning_mode,
+            memoized_flow_values_by_entity_name=self._active_model_memoized_values_by_entity_name,
+            memoized_query_values_by_entity_name={},
+            computed_descriptors=self._descriptors_computed_by_model,
+        )
         model_exception = None
         try:
-            model_value = self._compute_model_value(entity_name)
+            model_value = self._compute_model_value(entity_name, context)
         except CodeVersioningError as e:
             model_exception = e
 
@@ -128,6 +168,16 @@ class ModelFlowHarness:
         if model_exception is None:
             assert flow_value == model_value
 
+            if expect_exact_call_count_matches:
+                assert sorted(self._descriptors_computed_by_flow) == sorted(
+                    self._descriptors_computed_by_model
+                )
+            else:
+                assert set(self._descriptors_computed_by_flow) == set(
+                    self._descriptors_computed_by_model
+                )
+            self._clear_called_descriptors()
+
         else:
             assert flow_exception.__class__ == model_exception.__class__
 
@@ -156,100 +206,74 @@ class ModelFlowHarness:
                 update_version_annotation=True,
             )
 
-            self.query_and_check_entity(entity_name)
-
-        assert set(self._descriptors_computed_by_flow) == set(
-            self._descriptors_computed_by_model
-        )
-        self._clear_called_descriptors()
+            self.query_and_check_entity(
+                entity_name,
+                # Since our query was interrupted, some extra non-persisted entities
+                # may have been computed, so we can't expect the final counts to line
+                # up exactly.
+                expect_exact_call_count_matches=False,
+            )
 
     # This is just used for debugging.
     def save_dag(self, filename):
         self._builder.build().render_dag().save(filename)
 
+    def _clear_active_flow(self):
+        self._active_flow = None
+        self._active_model_memoized_values_by_entity_name = None
+
+    def _build_active_flow_if_missing(self):
+        if self._active_flow is None:
+            self._active_flow = self._builder.build()
+            self._active_model_memoized_values_by_entity_name = {}
+
     def _add_binding_to_flow(self, binding):
-        out_entity_names = [entity.name for entity in binding.out_entities]
+        binding.add_to_builder(self._builder, self._descriptors_computed_by_flow)
 
-        joined_dep_entity_names = ", ".join(binding.dep_entity_names)
-
-        # TODO Could this logic live inside ModelBinding instead?
-        if binding.use_tuples_for_output:
-            output_decorator_fragment = f"""
-            @bn.returns({binding.out_descriptor!r})
-            """.strip()
-            output_value_fragment = binding.value_code_fragment_for_returns()
-        else:
-            output_decorator_fragment = f"""
-            @bn.outputs({', '.join(repr(name) for name in out_entity_names)})
-            """.strip()
-            output_value_fragment = binding.value_code_fragment_for_outputs()
-
-        vars_dict = {
-            "bn": bn,
-            "builder": self._builder,
-            "record_call": self._descriptors_computed_by_flow.append,
-        }
-
-        raw_func_code = f"""
-        @builder
-        {output_decorator_fragment}
-        @bn.version({binding.annotated_func_version})
-        @bn.persist({binding.should_persist})
-        def _({joined_dep_entity_names}):
-            record_call({binding.out_descriptor!r})
-            return {output_value_fragment}
-        """
-        exec(dedent(raw_func_code), vars_dict)
-
-    def _compute_model_value(self, entity_name, memoized_values_by_entity_name=None):
-        if memoized_values_by_entity_name is None:
-            memoized_values_by_entity_name = {}
-        if entity_name in memoized_values_by_entity_name:
-            return memoized_values_by_entity_name[entity_name]
-
-        entity = self._entities_by_name[entity_name]
-        binding = entity.binding
-
-        if not binding.has_persisted_values:
-            dep_values = [
-                self._compute_model_value(dep_entity_name)
-                for dep_entity_name in binding.dep_entity_names
-            ]
-            values_by_entity_name = binding.compute_and_save_values(dep_values)
-            self._descriptors_computed_by_model.append(binding.out_descriptor)
-
-        else:
-            if binding.persisted_values_stale_ancestor is not None:
-                assert self._versioning_mode == "assist"
-                stale_ancestor = binding.persisted_values_stale_ancestor
-                raise CodeVersioningError("Binding is stale", stale_ancestor)
-            values_by_entity_name = binding.persisted_values_by_entity_name
-
-        memoized_values_by_entity_name.update(values_by_entity_name)
-        return values_by_entity_name[entity_name]
+    def _compute_model_value(self, entity_name, context):
+        binding = self._entities_by_name[entity_name].binding
+        return binding.compute_entity_value(entity_name, context)
 
     def _clear_called_descriptors(self):
         self._descriptors_computed_by_flow[:] = []
         self._descriptors_computed_by_model[:] = []
 
 
+@attr.s
+class ModelExecutionContext:
+    """
+    Represents the state associated with a modeled ``flow.get()`` call.
+    """
+
+    parallel_execution_enabled = attr.ib()
+    query_caching_enabled = attr.ib()
+    versioning_mode = attr.ib()
+    memoized_flow_values_by_entity_name = attr.ib()
+    memoized_query_values_by_entity_name = attr.ib()
+    computed_descriptors = attr.ib()
+
+    def evolve(self, **kwargs):
+        return attr.evolve(self, **kwargs)
+
+
 @attr.s
 class ModelEntity:
     """Represents an entity in a Bionic flow."""
 
     name = attr.ib()
     should_persist = attr.ib()
+    should_memoize = attr.ib()
 
     binding = attr.ib(default=None)
     dependent_entities = attr.ib(default=attr.Factory(list))
 
-    def compute_value(self, dep_values):
+    def value_from_deps(self, dep_values):
         return f"{self.name}.{self.binding.func_version}({','.join(dep_values)})"
 
     def value_code_fragment(self):
         return (
             f"'{self.name}.{self.binding.func_version}('"
-            + " + ',' ".join(f"+ {name}" for name in self.binding.dep_entity_names)
+            + " + ',' ".join(f"+ {entity.name}" for entity in self.binding.dep_entities)
             + " + ')'"
         )
 
@@ -262,7 +286,7 @@ class ModelBinding:
 
     out_dnode = attr.ib()
     out_entities = attr.ib()
-    dep_entity_names = attr.ib()
+    dep_entities = attr.ib()
     use_tuples_for_output = attr.ib()
     func_version = attr.ib(default=1)
     annotated_func_version = attr.ib(default=1)
@@ -355,9 +379,9 @@ class ModelBinding:
         else:
             assert False
 
-    def compute_and_save_values(self, dep_values):
+    def get_and_save_values_from_deps(self, dep_values):
         values_by_entity_name = {
-            entity.name: entity.compute_value(dep_values)
+            entity.name: entity.value_from_deps(dep_values)
             for entity in self.out_entities
         }
         if self.should_persist:
@@ -368,6 +392,82 @@ class ModelBinding:
         self.no_descendant_can_have_stale_persisted_values = False
         return values_by_entity_name
 
+    def compute_entity_value(self, entity_name, context):
+        if entity_name in context.memoized_query_values_by_entity_name:
+            return context.memoized_query_values_by_entity_name[entity_name]
+        if entity_name in context.memoized_flow_values_by_entity_name:
+            return context.memoized_flow_values_by_entity_name[entity_name]
+
+        if not self.has_persisted_values:
+            running_in_subprocess = (
+                context.parallel_execution_enabled and self.should_persist
+            )
+            if running_in_subprocess:
+                dep_context = context.evolve(
+                    memoized_flow_values_by_entity_name={},
+                    memoized_query_values_by_entity_name={},
+                )
+            else:
+                dep_context = context
+            dep_values = [
+                dep_entity.binding.compute_entity_value(dep_entity.name, dep_context)
+                for dep_entity in self.dep_entities
+            ]
+            values_by_entity_name = self.get_and_save_values_from_deps(dep_values)
+            context.computed_descriptors.append(self.out_descriptor)
+
+        else:
+            if self.persisted_values_stale_ancestor is not None:
+                assert context.versioning_mode == "assist"
+                stale_ancestor = self.persisted_values_stale_ancestor
+                raise CodeVersioningError("Binding is stale", stale_ancestor)
+            values_by_entity_name = self.persisted_values_by_entity_name
+
+        if self.should_memoize:
+            context.memoized_flow_values_by_entity_name.update(values_by_entity_name)
+        elif not self.should_persist and context.query_caching_enabled:
+            context.memoized_query_values_by_entity_name.update(values_by_entity_name)
+
+        return values_by_entity_name[entity_name]
+
+    def add_to_builder(self, builder, descriptors_computed_by_flow):
+        out_entity_names = [entity.name for entity in self.out_entities]
+
+        joined_dep_entity_names = ", ".join(self.dep_entity_names)
+
+        if self.use_tuples_for_output:
+            output_decorator_fragment = f"""
+            @bn.returns({self.out_descriptor!r})
+            """.strip()
+            output_value_fragment = self.value_code_fragment_for_returns()
+        else:
+            output_decorator_fragment = f"""
+            @bn.outputs({', '.join(repr(name) for name in out_entity_names)})
+            """.strip()
+            output_value_fragment = self.value_code_fragment_for_outputs()
+
+        vars_dict = {
+            "bn": bn,
+            "builder": builder,
+            "record_call": descriptors_computed_by_flow.append,
+        }
+
+        raw_func_code = f"""
+        @builder
+        {output_decorator_fragment}
+        @bn.version_no_warnings({self.annotated_func_version})
+        @bn.persist({self.should_persist})
+        @bn.memoize({self.should_memoize})
+        def _({joined_dep_entity_names}):
+            record_call({self.out_descriptor!r})
+            return {output_value_fragment}
+        """
+        import_code(dedent(raw_func_code), vars_dict=vars_dict)
+
+    @property
+    def dep_entity_names(self):
+        return [entity.name for entity in self.dep_entities]
+
     @property
     def out_descriptor(self):
         return self.out_dnode.to_descriptor()
@@ -378,6 +478,12 @@ class ModelBinding:
             entity.should_persist for entity in self.out_entities
         )
 
+    @property
+    def should_memoize(self):
+        return single_unique_element(
+            entity.should_memoize for entity in self.out_entities
+        )
+
 
 @attr.s
 class RandomFlowTestConfig:
@@ -393,6 +499,8 @@ class RandomFlowTestConfig:
         The number of update-and-query operations to perform on the flow.
     p_persist: float between 0 and 1
         The probability that any entity should be persistable.
+    p_memoize: float between 0 and 1
+        The probability that any entity should be memoizable.
     p_multi_out: float between 0 and 1
         The probability that any binding should output multiple entities.
     p_three_out_given_multi: float between 0 and 1
@@ -405,6 +513,9 @@ class RandomFlowTestConfig:
         If True, all definitions use ``@returns`` and may include nested tuples; if
         False, all definitions use ``@outputs`` and nested tuples are flattened into
         single-level tuples.
+    disable_query_caching: bool
+        If True, sets the ``core__temp_memoize_if_uncached`` entity in the tested flow
+        to False.
     versioning_mode: "manual", "assist", or "auto"
         Determines the value of ``core__versioning_mode`` in the tested flow.
     """
@@ -412,10 +523,12 @@ class RandomFlowTestConfig:
     n_bindings = attr.ib()
     n_iterations = attr.ib()
     p_persist = attr.ib(default=0.5)
+    p_memoize = attr.ib(default=0.5)
     p_multi_out = attr.ib(default=0.3)
     p_three_out_given_multi = attr.ib(default=0.3)
     p_update_version_annotation = attr.ib(default=1.0)
     use_tuples_for_output = attr.ib(default=True)
+    disable_query_caching = attr.ib(default=False)
     versioning_mode = attr.ib(default="manual")
 
 
@@ -445,19 +558,29 @@ class RandomFlowTester:
 
     def run(self):
         self._harness.set_versioning_mode(self._config.versioning_mode)
+        if self._config.disable_query_caching:
+            self._harness.disable_query_caching()
 
         for _ in range(self._config.n_bindings):
             self.add_random_binding()
         for _ in range(self._config.n_iterations):
             self.update_random_entity()
-            self.query_and_check_random_entity()
+            # Randomly run 0, 1, or 2 queries. This means we average 1 query per update,
+            # but sometimes perform multiple queries in a row (which is the only way
+            # we'll get to exercise the flow's in-memory cache).
+            for _ in range(self._rng.integers(3)):
+                self.query_and_check_random_entity()
 
     def add_random_binding(self):
         dep_entity_names = self._random_dep_entity_names()
         for dep_entity_name in dep_entity_names:
             self._child_counts_by_entity_name[dep_entity_name] += 1
         should_persist = self._random_bool(p_true=self._config.p_persist)
-        descriptor = self._random_descriptor_of_new_entities(should_persist)
+        should_memoize = self._random_bool(p_true=self._config.p_memoize)
+        descriptor = self._random_descriptor_of_new_entities(
+            should_persist,
+            should_memoize,
+        )
         self._harness.add_binding(
             out_descriptor=descriptor,
             dep_entity_names=dep_entity_names,
@@ -516,9 +639,9 @@ class RandomFlowTester:
     def _random_existing_entity_name(self):
         return self._rng.choice(self._harness.get_all_entity_names())
 
-    def _random_descriptor_of_new_entities(self, should_persist):
+    def _random_descriptor_of_new_entities(self, should_persist, should_memoize):
         def new_entity_name():
-            entity_name = self._harness.create_entity(should_persist)
+            entity_name = self._harness.create_entity(should_persist, should_memoize)
             self._child_counts_by_entity_name[entity_name] = 0
             return entity_name
 
@@ -548,8 +671,18 @@ class RandomFlowTester:
 
 
 @pytest.fixture
-def harness(builder, make_list):
-    return ModelFlowHarness(builder, make_list)
+def harness(builder, make_list, parallel_execution_enabled):
+    return ModelFlowHarness(
+        builder,
+        make_list,
+        parallel_execution_enabled=parallel_execution_enabled,
+    )
+    # If parallel execution is enabled, some non-persistable entities may get
+    # computed multiple times in different processes, which our model doesn't
+    # account for.
+    if parallel_execution_enabled:
+        harness.disable_exact_call_counting()
+    return harness
 
 
 # This test is mostly here as an example of how to use the harness in a deterministic
@@ -578,14 +711,18 @@ def test_random_flow_small(harness):
 
 
 @pytest.mark.slow
-@pytest.mark.parametrize("p_persist", [0.2, 0.5, 0.8])
+@pytest.mark.parametrize("p_persist", [0.3, 0.7])
+@pytest.mark.parametrize("p_memoize", [0.3, 0.7])
 @pytest.mark.parametrize("p_multi_out", [0.2, 0.6])
 @pytest.mark.parametrize("n_bindings, n_iterations", [(10, 20), (50, 50)])
-def test_random_flow_varied(harness, n_bindings, n_iterations, p_persist, p_multi_out):
+def test_random_flow_varied(
+    harness, n_bindings, n_iterations, p_persist, p_memoize, p_multi_out
+):
     config = RandomFlowTestConfig(
         n_bindings=n_bindings,
         n_iterations=n_iterations,
         p_persist=p_persist,
+        p_memoize=p_memoize,
         p_multi_out=p_multi_out,
     )
     RandomFlowTester(harness, config).run()
@@ -622,3 +759,24 @@ def test_random_flow_no_tuples(harness):
         use_tuples_for_output=False,
     )
     RandomFlowTester(harness, config).run()
+
+
+@pytest.mark.slow
+def test_random_flow_no_query_caching(harness):
+    # Unfortunately, when query caching is disabled, it's difficult to model the
+    # exact number of times a function will be called. The main problem is when we have
+    # a tuple `x, y` and we need both `x` and `y` in different parts of the flow. If
+    # both requirements are detected before we compute `x, y`, then both entities will
+    # be stored in memory until they can be used. Otherwise, the non-required entity
+    # will be evicted and `x, y` will have to be recomputed when the second requirement
+    # is detected.
+    # In the future we may be able to refactor our implementation to be more
+    # deterministic (i.e., always discard any values that aren't needed immediately),
+    # in which case we can try to model its behavior more closely.
+    harness.disable_exact_call_counting()
+    config = RandomFlowTestConfig(
+        n_bindings=50,
+        n_iterations=50,
+        disable_query_caching=True,
+    )
+    RandomFlowTester(harness, config).run()
